import re

from dateutil.parser import parse
from dateutil.relativedelta import relativedelta

re_iso_date = re.compile(r"^\d\d\d\d-\d\d-\d\d$")
re_iso_time = re.compile(r"^\d\d:\d\d:\d\d(\.\d\d\d\d\d\d)?$")
re_iso_datetime = re.compile(r"^\d\d\d\d-\d\d-\d\dT\d\d:\d\d:\d\d(\.\d\d\d\d\d\d)?$")


def get_datetime_dict(type, value_1=None, value_2=None):
    """Helper to create the dict structure for DateTime."""
    d = {"_type": "DateTime", "type": type}

    if type in ["datetime", "time", "date"]:
        d["value"] = value_1
    else:
        d["value"] = f"{value_1} to {value_2}"
        if value_1:
            d["from_value"] = {"type": type.split("_")[0], "value": value_1}
        if value_2:
            d["to_value"] = {"type": type.split("_")[0], "value": value_2}

    return d


def convert(value: str, type_name: str):
    """Helper to convert a value to a new type.

    Initial use case, to convert string to DateTime objects.
    """
    if type_name in ["time", "date", "datetime"]:
        return get_datetime_dict(type_name, value)

    # Otherwise, just return the value
    return value


def datetime(*args):
    """Constructor for a datetime object."""
    if len(args) == 1:
        return get_datetime_dict("datetime", args[0])
    elif len(args) == 2:
        # We're dealing with a date and a time
        date_str = args[0]
        if isinstance(date_str, dict):
            date_str = date_str["value"]

        time_str = args[1]
        if isinstance(time_str, dict):
            time_str = time_str["value"]

        return get_datetime_dict("datetime", f"{date_str}T{time_str}")

    return None


def datetime_add(val, hours=None, days=None, weeks=None, months=None, years=None):
    """Adds the specified components to a datetime."""
    if isinstance(val, str):
        dt_val = parse(val)
    elif isinstance(val, dict) and val["type"] in ["date", "datetime"]:
        dt_val = parse(val["value"])
    else:
        raise Exception(f"Unknown datetime format: {val}")

    delta = relativedelta(
        hours=hours or 0,
        days=days or 0,
        weeks=weeks or 0,
        months=months or 0,
        years=years or 0,
    )
    dt_val = dt_val + delta

    return get_datetime_dict("datetime", dt_val.isoformat())


def date_add(val, days=None, weeks=None, months=None, years=None):
    """Adds the specified components to a date."""
    result = datetime_add(val, days=days, weeks=weeks, months=months, years=years)
    return get_datetime_dict("date", result["value"][0:10])


def date(*args):
    """Constructor for a date.

    Works both with a string and a DateTime instance.
    """
    if not args:
        return None

    if isinstance(args[0], str):
        if re_iso_date.match(args[0]) or re_iso_datetime.match(args[0]):
            return get_datetime_dict("date", args[0][0:10])

    elif isinstance(args[0], dict):
        if args[0]["type"] == "datetime":
            return get_datetime_dict("date", args[0]["value"][0:10])
        elif args[0]["type"] == "date":
            return get_datetime_dict("date", args[0]["value"])

    return None


def time(*args):
    """Constructor for a time.

    Works both with a string and a DateTime instance.
    """
    if not args:
        return None

    if isinstance(args[0], str):
        if re_iso_time.match(args[0]):
            return get_datetime_dict("time", args[0])
        elif re_iso_datetime.match(args[0]):
            return get_datetime_dict("time", args[0][11:])

    elif isinstance(args[0], dict):
        if args[0]["type"] == "datetime":
            return get_datetime_dict("time", args[0]["value"][11:])

    return None


def datetime_interval(*args):
    """Constructor for a datetime_interval object"""
    if len(args) == 2:
        if (
            isinstance(args[0], dict)
            and isinstance(args[1], dict)
            and args[0]["type"] == "date"
            and args[1]["type"] == "time_interval"
        ):
            from_value = f"{args[0]['value']}T{args[1]['from_value']['value']}"
            to_value = f"{args[0]['value']}T{args[1]['to_value']['value']}"
            return get_datetime_dict("datetime_interval", from_value, to_value)

        elif (
            isinstance(args[0], dict)
            and isinstance(args[1], dict)
            and args[0]["type"] == "datetime"
            and args[1]["type"] == "datetime"
        ):
            return get_datetime_dict(
                "datetime_interval", args[0]["value"], args[1]["value"]
            )

    return None


def weekday(dt):
    res = parse(dt["value"]).weekday()
    return res


# The list of default functions available when evaluating expressions
default_functions = {
    "convert": convert,
    "date": date,
    "time": time,
    "datetime": datetime,
    "datetime_interval": datetime_interval,
    "date_add": date_add,
    "datetime_add": datetime_add,
    "weekday": weekday,
}



from typing import Optional

from pydantic import BaseModel


class ConversationData(BaseModel):
    bot_id: str
    user_id: Optional[str] = None
    authenticated_user_id: Optional[str] = None
    webhook_url: Optional[str] = None
    conversation_id: Optional[str] = None
    context_data: Optional[dict] = None


class EventData(BaseModel):
    event_type: str
    event_data: dict


class MessageData(BaseModel):
    message: str


# We have to use the below hack because "from" is a keyword in python
MessageData.__annotations__["from"] = str



import logging
from typing import Any, Text, Dict, List

from rasa_sdk import Action
from rasa_sdk import Tracker
from rasa_sdk.events import SlotSet
from rasa_sdk.executor import CollectingDispatcher

from .util import eval_expr

log = logging.getLogger(__name__)


class ActionContext:
    """Context object that is passed to actions."""
    def __init__(self, dispatcher: CollectingDispatcher,
                 tracker: Tracker,
                 domain: Dict[Text, Any]):
        self.dispatcher = dispatcher
        self.tracker = tracker
        self.domain = domain

        # We keep track of all the new slots that are set
        self.action_events = []
        self.new_slots = {}

    async def get(self, key, default_value=None):
        """Returns the value of a slot."""
        if key in self.new_slots:
            value = self.new_slots[key]
        else:
            value = self.tracker.get_slot(key)

        if value is None:
            value = default_value

        return value

    async def set(self, key, value):
        """Sets the value of a slot."""
        self.new_slots[key] = value

    async def update(self, d):
        """Updates the context by setting the slots specified in the dict."""
        self.new_slots.update(d)

    def add_event(self, event: any):
        self.action_events.append(event)

    def get_slot_set_events(self):
        """Computes the slot_set event with all the values."""

        events = []
        for k, v in self.new_slots.items():
            events.append(SlotSet(k, v))

        return events


def action(action_name: str):
    """Decorator for registering an action with the specified name.

    The parameters of the action are loaded from the `action_params` slot.
    There is basic support to fetch the value of some parameters from other context
    variables.

    The return value of the action is set into the `action_result`.

    :param action_name: The name under which the action should be registered.
    """

    def decorator(func):
        log.debug(f"Registering action '{action_name}' ...")

        action_class = type(f"Action_{action_name}", (Action,), {})

        def name(self) -> Text:
            return action_name

        action_class.name = name

        async def run(self, dispatcher: CollectingDispatcher,
                      tracker: Tracker,
                      domain: Dict[Text, Any]) -> List[Dict[Text, Any]]:

            # We extract the action params
            action_params = tracker.get_slot("action_params") or {}

            # We go through them and if there are references to other slots, we fill them in
            for key, value in list(action_params.items()):
                if isinstance(value, str) and value[0] == "$" and value[1] != "$":
                    action_params[key] = eval_expr(tracker, value)

            # We invoke the action
            ctx = ActionContext(dispatcher, tracker, domain)
            action_result = await func(ctx, **action_params)

            ctx.new_slots["action_result"] = action_result

            # If there were any slots set, we also generate events for them.
            slot_set_events = ctx.get_slot_set_events()

            return slot_set_events + ctx.action_events

        action_class.run = run

        return action_class

    return decorator



import json
import logging
import re
from datetime import datetime
from typing import Optional, List

from dateutil import parser
from dateutil import parser as date_parser
from dateutil.relativedelta import relativedelta

from ..actions.default_functions import get_datetime_dict

logger = logging.getLogger(__name__)

re_iso_date = re.compile(r"^\d\d\d\d-\d\d-\d\d$")
re_iso_time = re.compile(r"^\d\d:\d\d:\d\d(?:\.\d\d\d\d\d\d)?$")
re_iso_datetime = re.compile(r"^\d\d\d\d-\d\d-\d\dT\d\d:\d\d:\d\d(?:\.\d\d\d\d\d\d)?$")

DATE_WORDS = [
    "now",
    # relative days
    "today",
    "tomorrow",
    "yesterday",
    # weekdays
    "monday",
    "tuesday",
    "wednesday",
    "thursday",
    "friday",
    "saturday",
    "sunday",
    # months
    "january",
    "jan",
    "february",
    "feb",
    "march",
    "mar",
    "april",
    "apr",
    "may",
    "june",
    "jun",
    "july",
    "jul",
    "august",
    "aug",
    "september",
    "sep",
    "october",
    "oct",
    "november",
    "nov",
    "december",
    "dec",
    # periods
    "week",
    "month",
]
re_date_mention = re.compile(
    "|".join([rf"\b{word}\b" for word in DATE_WORDS]), re.IGNORECASE
)
re_year_mention = re.compile(r"\d\d\d\d")
re_month_year_mention = re.compile(r"\d\d[-/.]\d\d")

TIME_WORDS = [
    "now",
    # periods in the day
    "afternoon",
    "morning",
    "evening",
    "noon",
    # interval markers
    "after",
    "before",
]
re_time_mention = re.compile(
    "|".join([rf"\b{word}\b" for word in TIME_WORDS]), re.IGNORECASE
)
re_time_am_pm = re.compile(r"\d *[ap]m", re.IGNORECASE)
re_time = re.compile(r"\d:\d", re.IGNORECASE)


def has_date_mentioned(text: str):
    """Helper to determine if a text mentions something that looks like a date."""
    return (
        len(re_date_mention.findall(text)) > 0
        or len(re_year_mention.findall(text)) > 0
        or len(re_month_year_mention.findall(text)) > 0
    )


def has_time_mentioned(text: str):
    """Helper to determine if a text mentions something that looks like a time."""
    return (
        len(re_time_mention.findall(text)) > 0
        or len(re_time_am_pm.findall(text)) > 0
        or len(re_time.findall(text)) > 0
    )


def parse_duckling_datetime(res: dict):
    """Parses the provided text using the duckling data from the context."""
    if res["dim"] != "time":
        return res

    # We only look at the first value for now
    value = res["value"]

    # Check whether we have date / time mentions to help us decide on
    # the type of value we are dealing with
    has_date = has_date_mentioned(res["body"])

    if value["type"] == "value":
        time_string = value["value"]
        grain = value["grain"]
        datetime_val_tz = date_parser.parse(time_string)
        datetime_val = datetime_val_tz.replace(tzinfo=None)

        # We have to decide what type of value we're dealing with
        if grain in ["second", "minute", "hour"]:
            # If we have an time component, we need to check if there's a date component too
            if has_date:
                return get_datetime_dict("datetime", datetime_val.isoformat())
            else:
                return get_datetime_dict("time", datetime_val.isoformat()[11:])
        elif grain == "day":
            # we're dealing with a day
            return get_datetime_dict("date", datetime_val.strftime("%Y-%m-%d"))
        elif grain in ["week", "month", "year"]:
            # we're dealing with a multi-day interval
            if grain == "week":
                delta = relativedelta(days=7)
            elif grain == "month":
                delta = relativedelta(months=1)
            else:
                delta = relativedelta(years=1)

            datetime_end_val = datetime_val + delta
            start_date = datetime_val.strftime("%Y-%m-%d")
            end_date = datetime_end_val.strftime("%Y-%m-%d")

            return get_datetime_dict("date_interval", start_date, end_date)
        else:
            logger.error(f"Found unknown duckling grain in: {json.dumps(value)}")

    elif value["type"] == "interval":
        # we work only with the grain of the "from" component (or the "to" if it's the only one)
        grain = (
            value["from"]["grain"] if value.get("from") else value["to"]["grain"]
        )

        datetime_from = (
            date_parser.parse(value["from"]["value"]).replace(tzinfo=None)
            if value.get("from")
            else None
        )
        datetime_to = (
            date_parser.parse(value["to"]["value"]).replace(tzinfo=None)
            if value.get("to")
            else None
        )

        # Tweak to the evening interval, make it 6pm to 11pm (rather than midnight)
        if datetime_to and datetime_from:
            if (
                datetime_to - datetime_from
            ).total_seconds() == 60 * 60 * 6 and datetime_from.hour == 18:
                datetime_to = datetime_from + relativedelta(hours=5)

        if grain in ["second", "minute", "hour"]:
            if not has_date:
                start_time = (
                    datetime_from.isoformat()[11:] if datetime_from else None
                )
                end_time = datetime_to.isoformat()[11:] if datetime_to else None

                return get_datetime_dict("time_interval", start_time, end_time)
            else:
                start_datetime = (
                    datetime_from.isoformat() if datetime_from else None
                )
                end_datetime = datetime_to.isoformat() if datetime_to else None

                return get_datetime_dict(
                    "datetime_interval", start_datetime, end_datetime
                )
        else:
            # Otherwise it's a date interval
            start_date = datetime_from.isoformat()[11:] if datetime_from else None
            end_date = datetime_to.isoformat()[11:] if datetime_to else None

            return get_datetime_dict("date_interval", start_date, end_date)
    else:
        raise Exception(f"Unknown duckling value type: {value['type']}")

    return None


def date_to_str(dt, value_format: str):
    dt_val = parser.parse(dt["value"])
    if value_format == "friendly":
        weekday = dt_val.strftime("%A")
        month = dt_val.strftime("%B")

        day = str(dt_val.day)
        if day[-1] == "1" and (len(day) == 1 or day[-2] != "1"):
            day += "st"
        elif day[-1] == "2" and (len(day) == 1 or day[-2] != "1"):
            day += "nd"
        elif day[-1] == "3" and (len(day) == 1 or day[-2] != "1"):
            day += "rd"
        else:
            day += "th"

        # e.g. "Friday, 5th October"
        return f"{weekday}, {month} {day}"

    elif value_format == "relative":
        pass

    return dt["value"]


def time_to_str(dt, value_format: str):
    try:
        dt_val = datetime.strptime(dt["value"], "%H:%M:%S.%f")
    except ValueError:
        dt_val = datetime.strptime(dt["value"], "%H:%M:%S")

    if value_format == "friendly":
        hour = dt_val.hour
        am_pm = "am"
        if hour >= 12:
            am_pm = "pm"
            if hour > 12:
                hour -= 12

        # We want 12am not 0am
        if hour == 0:
            hour = 12

        if dt_val.minute > 0:
            return f"{hour}:{dt_val.minute:02d}{am_pm}"
        else:
            res = f"{hour}{am_pm}"
            if res == "12pm":
                return "noon"
            elif res == "12am":
                return "midnight"
            else:
                return res

    elif value_format == "relative":
        pass

    return dt_val.strftime("%H:%M:%S")


def datetime_to_str(dt, value_format: str):
    date_part = date_to_str({"type": "date", "value": dt["value"][0:10]}, value_format)

    time_part = time_to_str({"type": "time", "value": dt["value"][11:]}, value_format)

    return date_part + " at " + time_part


def date_interval_to_str(dt, value_format: str):
    if dt["from_value"] and dt["to_value"]:
        return (
            "from "
            + date_to_str(dt["from_value"], value_format)
            + " until "
            + date_to_str(dt["to_value"], value_format)
        )
    elif dt["from_value"]:
        return "from " + date_to_str(dt["from_value"], value_format)
    else:
        return "until " + date_to_str(dt["to_value"], value_format)


def time_interval_to_str(dt, value_format: str):
    if dt["from_value"] and dt["to_value"]:
        res = (
            "from "
            + time_to_str(dt["from_value"], value_format)
            + " until "
            + time_to_str(dt["to_value"], value_format)
        )

        if value_format == "friendly":
            if res == "from 4am until noon":
                return "in the morning"
            elif res == "from noon until 7pm":
                return "in the afternoon"
            elif res == "from 6pm until 11pm":
                return "evening"

        return res

    elif dt["from_value"]:
        return "from " + time_to_str(dt["from_value"], value_format)
    else:
        return "until " + time_to_str(dt["to_value"], value_format)


def datetime_interval_to_str(dt, value_format: str):
    if dt.get("from_value") and dt.get("to_value"):
        # If it's the same day, we use the time interval
        if dt["from_value"]["value"][0:10] == dt["to_value"]["value"][0:10]:
            from_time = dt["from_value"]["value"][11:]
            to_time = dt["to_value"]["value"][11:]

            return (
                date_to_str(dt["from_value"], value_format)
                + ", "
                + time_interval_to_str(
                    {
                        "type": "time_interval",
                        "value": f"{from_time} to {to_time}",
                        "from_value": {"type": "time", "value": from_time},
                        "to_value": {"type": "time", "value": to_time},
                    },
                    value_format,
                )
            )
        else:
            return (
                "from "
                + datetime_to_str(dt["from_value"], value_format)
                + " until "
                + datetime_to_str(dt["to_value"], value_format)
            )

    elif dt.get("from_value"):
        return "from " + datetime_to_str(dt["from_value"], value_format)
    else:
        return "until " + datetime_to_str(dt["to_value"], value_format)


def _format_datetime(dt: dict, value_format: str):
    if value_format in ["default", "relative", "friendly"]:
        if dt["type"] == "date":
            return date_to_str(dt, value_format)
        elif dt["type"] == "time":
            return time_to_str(dt, value_format)
        elif dt["type"] == "datetime":
            return datetime_to_str(dt, value_format)
        if dt["type"] == "date_interval":
            return date_interval_to_str(dt, value_format)
        elif dt["type"] == "time_interval":
            return time_interval_to_str(dt, value_format)
        elif dt["type"] == "datetime_interval":
            return datetime_interval_to_str(dt, value_format)
    elif value_format in ["date"]:
        # We need to extract the date from a datetime
        return get_datetime_dict("date", dt["value"][0:10])
    elif value_format in ["time"]:
        # We need to extract the time from a datetime
        return get_datetime_dict("time", dt["value"][11:])

    # For any cases not caught
    return dt["value"]


def nlg_formatter(value: any, value_formats: Optional[List[str]] = None):
    """A more advanced formatter that can deal with objects, DateTime and multi-language.

    :param value: The value to be formatted.
    :param value_formats: A list of strings describing the formats to use for the output.
    """
    if value_formats is None or len(value_formats) == 0:
        value_formats = ["default"]

    for value_format in value_formats:
        # We try to detect automatically certain types
        if isinstance(value, str):
            if re_iso_date.match(value):
                value = get_datetime_dict("date", value)
            elif re_iso_time.match(value):
                value = get_datetime_dict("time", value)
            elif re_iso_datetime.match(value):
                value = get_datetime_dict("datetime", value)

        if isinstance(value, dict):
            if value.get("_type") == "DateTime":
                value = _format_datetime(value, value_format)
            elif isinstance(value, dict) and "name" in value:
                value = value["name"]

    return value



from datetime import datetime, timezone
from typing import Any, Text, Dict, List

import rasa_sdk.events
from .util import eval_expr
from rasa_sdk import Action, Tracker
from rasa_sdk.events import SlotSet
from rasa_sdk.executor import CollectingDispatcher


class InitializeConversation(Action):

    def name(self) -> Text:
        return 'initialize_conversation'

    def run(self, dispatcher: CollectingDispatcher,
            tracker: Tracker,
            domain: Dict[Text, Any]) -> List[Dict[Text, Any]]:

        events = []

        # We read the information from the "new_conversation" event
        event = tracker.get_slot("event_data")

        context_updates = {}
        if "bot_id" in event:
            context_updates["config.live_bot_name"] = event["bot_id"]
        if "user_id" in event:
            context_updates["user_id"] = event["user_id"]
        if "authenticated_user_id" in event:
            context_updates["authenticated_user_id"] = event["authenticated_user_id"]
        if "context_data" in event and isinstance(event["context_data"], dict):
            context_updates.update(event["context_data"])

        for k, v in context_updates.items():
            events.append(SlotSet(k, v))

        # We also set the current date and time for the conversation
        now = datetime.now(timezone.utc).astimezone().isoformat()
        events.append(SlotSet("system_datetime", now[0:26]))
        events.append(SlotSet("system_date", now[0:10]))
        events.append(SlotSet("system_time", now[11:26]))

        # We set the `eval` slot to the truthful value of the expression.
        events.extend([
            rasa_sdk.events.ActionExecuted(action_name="action_listen"),
            rasa_sdk.events.UserUttered(
                text="conversation_initialized",
                parse_data={
                    "intent": {
                        "name": "event_conversation_initialized",
                        "confidence": 1.0
                    },
                    "entities": []
                }
            )
        ])

        return events


class HangUp(Action):

    def name(self) -> Text:
        return 'hang_up'

    def run(self, dispatcher: CollectingDispatcher,
            tracker: Tracker,
            domain: Dict[Text, Any]) -> List[Dict[Text, Any]]:

        return []


class Eval(Action):

    def name(self) -> Text:
        return 'eval'

    def run(self, dispatcher: CollectingDispatcher,
            tracker: Tracker,
            domain: Dict[Text, Any]) -> List[Dict[Text, Any]]:

        if not hasattr(tracker, "action_parameters"):
            raise Exception("No action parameters found in tracker.")

        expr = getattr(tracker, "action_parameters")
        result = eval_expr(tracker, expr)

        # We set the `eval` slot to the truthful value of the expression.
        return [
            SlotSet(key="eval", value=True if result else False),
        ]


class Set(Action):

    def name(self) -> Text:
        return 'set'

    def run(self, dispatcher: CollectingDispatcher,
            tracker: Tracker,
            domain: Dict[Text, Any]) -> List[Dict[Text, Any]]:

        if not hasattr(tracker, "action_parameters"):
            raise Exception("No action parameters found in tracker.")

        var_name, _, expr = getattr(tracker, "action_parameters").partition("=")
        var_name = var_name.strip().replace("$", "")
        val = eval_expr(tracker, expr)

        return [
            SlotSet(key=var_name, value=val),
        ]


class Push(Action):

    def name(self) -> Text:
        return 'push'

    def run(self, dispatcher: CollectingDispatcher,
            tracker: Tracker,
            domain: Dict[Text, Any]) -> List[Dict[Text, Any]]:

        if not hasattr(tracker, "action_parameters"):
            raise Exception("No action parameters found in tracker.")

        return_stack = tracker.slots.get("return_stack")
        assert isinstance(return_stack, list)

        return_point = getattr(tracker, "action_parameters")
        return_stack.append(return_point)

        print("Updated return stack to: ", return_stack)

        return [
            SlotSet(key="return_stack", value=return_stack),
        ]


class Call(Action):

    def name(self) -> Text:
        return 'call'

    def run(self, dispatcher: CollectingDispatcher,
            tracker: Tracker,
            domain: Dict[Text, Any]) -> List[Dict[Text, Any]]:

        return []


class Pop(Action):

    def name(self) -> Text:
        return 'pop'

    def run(self, dispatcher: CollectingDispatcher,
            tracker: Tracker,
            domain: Dict[Text, Any]) -> List[Dict[Text, Any]]:

        return_stack = tracker.slots.get("return_stack")
        assert isinstance(return_stack, list)

        # If we have action parameters, we're dealing with a specific pop, and we don't
        # need to do anything.
        if not hasattr(tracker, "action_parameters"):
            # If we have something left in the return stack, we use it.
            if return_stack:
                return_point = return_stack.pop()

                return [
                    SlotSet(key="return_stack", value=return_stack),
                    rasa_sdk.events.FollowupAction(f"pop({return_point})")
                ]

        return []


class CreateEvent(Action):

    def name(self) -> Text:
        return 'create_event'

    def run(self, dispatcher: CollectingDispatcher,
            tracker: Tracker,
            domain: Dict[Text, Any]) -> List[Dict[Text, Any]]:

        entities = []

        # Support for `create_event(bla)` with the event type specified inline
        if hasattr(tracker, "action_parameters"):
            # TODO: add support for additional event parameters.
            event_type = getattr(tracker, "action_parameters")
        else:
            action_parameters = tracker.slots.get("action_params", {})
            event = action_parameters.get("event", {})
            event_type = event["_type"]
            for k, v in event.items():
                if k.startswith("_"):
                    continue

                entities.append({
                    "entity": k,
                    "start": 0,
                    "end": 0,
                    "value": v
                })

        return [
            rasa_sdk.events.ActionExecuted(action_name="action_listen"),
            rasa_sdk.events.UserUttered(
                text=event_type,
                parse_data={
                    "intent": {
                        "name": f"event_{event_type}",
                        "confidence": 1.0
                    },
                    "entities": entities
                }
            )
        ]



from typing import Dict, Text, Any
from typing import List

import rasa.shared.utils.io
from rasa.nlu.extractors.duckling_entity_extractor import DucklingEntityExtractor as RASADucklingEntityExtractor
from rasa.shared.constants import DOCS_URL_COMPONENTS
from rasa.shared.nlu.constants import ENTITIES, TEXT
from rasa.shared.nlu.training_data.message import Message
from .datetime_util import parse_duckling_datetime


def extract_value(match: Dict[Text, Any]) -> Dict[Text, Any]:
    if match["value"].get("type") == "interval":
        value = {
            "to": match["value"].get("to", {}).get("value"),
            "from": match["value"].get("from", {}).get("value"),
        }
    else:
        value = match["value"].get("value")

    return value


def convert_duckling_format_to_rasa(
    matches: List[Dict[Text, Any]]
) -> List[Dict[Text, Any]]:
    extracted = []

    for match in matches:
        if match["dim"] == "time":
            entity = "datetime"
            value = parse_duckling_datetime(match)
        else:
            entity = match["dim"]
            value = extract_value(match)

        entity = {
            "start": match["start"],
            "end": match["end"],
            "text": match.get("body", match.get("text", None)),
            "value": value,
            "confidence": 1.0,
            "additional_info": match["value"],
            "entity": entity,
        }

        extracted.append(entity)

    return extracted


class DucklingEntityExtractor(RASADucklingEntityExtractor):
    """Wrapper on top of the original RASA entity extractor."""

    def process(self, message: Message, **kwargs: Any) -> None:

        if self._url() is not None:
            reference_time = self._reference_time_from_message(message)
            matches = self._duckling_parse(message.get(TEXT), reference_time)
            all_extracted = convert_duckling_format_to_rasa(matches)
            dimensions = self.component_config["dimensions"]
            extracted = DucklingEntityExtractor.filter_irrelevant_entities(
                all_extracted, dimensions
            )
        else:
            extracted = []
            rasa.shared.utils.io.raise_warning(
                "Duckling HTTP component in pipeline, but no "
                "`url` configuration in the config "
                "file nor is `RASA_DUCKLING_HTTP_URL` "
                "set as an environment variable. No entities will be extracted!",
                docs=DOCS_URL_COMPONENTS + "#DucklingEntityExtractor",
            )

        extracted = self.add_extractor_name(extracted)
        message.set(ENTITIES, message.get(ENTITIES, []) + extracted, add_to_output=True)



""" In order to add support for actions to have parameters patch the ActionExecutor.run function."""
import logging
import typing
from typing import Text, Dict, Any, Optional

from rasa_sdk import utils
from rasa_sdk.executor import CollectingDispatcher
from rasa_sdk.interfaces import ActionNotFoundException

if typing.TYPE_CHECKING:  # pragma: no cover
    from rasa_sdk.types import ActionCall

logger = logging.getLogger(__name__)


async def run(self, action_call: "ActionCall") -> Optional[Dict[Text, Any]]:
    from rasa_sdk.interfaces import Tracker

    action_name = action_call.get("next_action")
    if action_name:
        logger.debug(f"Received request to run '{action_name}'")

        # If the action name has the form "name(parameters)" we extract them.
        parameters = None
        if "(" in action_name:
            action_name, _, parameters = action_name.partition("(")
            parameters = parameters.strip()
            assert parameters[-1] == ")"
            parameters = parameters[0:-1]

        action = self.actions.get(action_name)
        if not action:
            raise ActionNotFoundException(action_name)

        tracker_json = action_call["tracker"]
        domain = action_call.get("domain", {})
        tracker = Tracker.from_dict(tracker_json)
        dispatcher = CollectingDispatcher()

        if parameters:
            tracker.action_parameters = parameters

        events = await utils.call_potential_coroutine(
            action(dispatcher, tracker, domain)
        )

        if not events:
            # make sure the action did not just return `None`...
            events = []

        validated_events = self.validate_events(events, action_name)
        logger.debug(f"Finished running '{action_name}'")
        return self._create_api_response(validated_events, dispatcher.messages)

    logger.warning("Received an action call without an action.")
    return None



import asyncio
import json
import logging
import os
import secrets
from time import time
from typing import Optional

from fastapi import FastAPI, Depends, HTTPException
from fastapi import Response, Form, Request
from fastapi.security import HTTPBasic, HTTPBasicCredentials
from starlette import status
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware

from . import rasa_api
from .models import ConversationData, EventData, MessageData
from .twilio_api import start_recording

log = logging.getLogger(__name__)

middleware = [
    Middleware(
        CORSMiddleware,
        allow_origins=['*'],
        allow_credentials=True,
        allow_methods=['*'],
        allow_headers=['*']
    )
]

app = FastAPI(middleware=middleware)

security = HTTPBasic()

# Fetch the HTTP authentication information
BASIC_HTTP_AUTH = os.environ.get("BASIC_HTTP_AUTH")
CREDENTIALS = [item.split(":") for item in BASIC_HTTP_AUTH.split(",")]

WEBCHAT_KEY = [cred[1] for cred in CREDENTIALS
               if cred[0] == "webchat"][0]


def get_auth_key(credentials: HTTPBasicCredentials = Depends(security)):
    found = False

    for cred in CREDENTIALS:
        correct_username = secrets.compare_digest(credentials.username, cred[0])
        correct_password = secrets.compare_digest(credentials.password, cred[1])
        if correct_username and correct_password:
            found = True
            break

    if not found:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
            headers={"WWW-Authenticate": "Basic"},
        )

    return credentials.username


_bot_id_cache = {}


async def _get_bot_id(conversation_id: str):
    """Returns the bot_id associated with a conversation."""
    tries = 0

    # We wait up to 5 seconds for it to appear in the cache
    while True:
        if conversation_id not in _bot_id_cache:
            if tries < 10:
                tries += 1
                await asyncio.sleep(0.5)
            else:
                raise Exception(f"Bot id for conversation {conversation_id} unknown.")

                # TODO: save the cache in a permanent storage so that it works with multiple
                #   instances.
        else:
            break

    return _bot_id_cache[conversation_id]


@app.get("/")
async def root():
    return {"status": "ok"}


@app.post("/conversations")
async def create_conversation(conversation_data: ConversationData,
                              request: Request,
                              auth_key: str = Depends(get_auth_key)):
    """Creates a new conversation."""
    log.info(f"CREATE_CONVERSATION [{auth_key}] {request.url}")
    log.info(await request.body())

    conversation_id, events = await rasa_api.create_conversation(
        conversation_data.bot_id, conversation_data.dict()
    )

    # Update the cache
    _bot_id_cache[conversation_id] = conversation_data.bot_id

    return {
        "conversation_id": conversation_id,
        "status": "active",
        "participants": {
            "user": "",
            "bot": conversation_data.bot_id
        },
        "created_at": events[0]["created_at"]
    }


@app.post("/conversations/{conversation_id}/events")
async def add_event(conversation_id: str,
                    event_data: EventData,
                    request: Request,
                    wait_response: bool = False,
                    auth_key: str = Depends(get_auth_key)):
    """Adds a new event to a conversation."""
    log.info(f"ADD_EVENT [{auth_key}] {request.url}")
    log.info(await request.body())

    bot_id = await _get_bot_id(conversation_id)

    initial_events = await rasa_api.add_event(bot_id, conversation_id, event_data.dict())

    # If we need to wait for response, we run the reasoning
    if wait_response:
        events = await rasa_api.run_reasoning(bot_id, conversation_id)
    else:
        events = initial_events

    # Return the new events including the new one
    return list(reversed(events[len(initial_events) - 1:]))


@app.post("/conversations/{conversation_id}/messages")
async def add_message(conversation_id: str,
                      message_data: MessageData,
                      request: Request,
                      wait_response: bool = False,
                      auth_key: str = Depends(get_auth_key)):
    """Adds a new message to a conversation."""
    log.info(f"ADD_MESSAGE [{auth_key}] {request.url}")
    log.info(await request.body())

    bot_id = await _get_bot_id(conversation_id)

    raw_data = json.loads(request._body)
    initial_events = await rasa_api.add_event(bot_id, conversation_id, {
        "event_type": "user_said" if raw_data.get("from") == "user" else "bot_said",
        "event_data": {
            "value": message_data.message
        }
    })

    # If we need to wait for response, we run the reasoning
    if wait_response:
        events = await rasa_api.run_reasoning(bot_id, conversation_id)
    else:
        events = initial_events

    # Return the new events including the new one
    return list(reversed(events[len(initial_events) - 1:]))


@app.get("/conversations/{conversation_id}/events")
async def list_events(conversation_id: str,
                      request: Request,
                      wait_on_event_id: Optional[str] = None,
                      auth_key: str = Depends(get_auth_key)):
    log.info(f"[{auth_key}] {request.url}")
    log.info(await request.body())

    bot_id = await _get_bot_id(conversation_id)

    t0 = time()
    events = await rasa_api.list_events(bot_id, conversation_id)

    while (wait_on_event_id and events
            and events[-1]["event_id"] == wait_on_event_id and time() - t0 < 25):
        await asyncio.sleep(1)

        events = await rasa_api.list_events(bot_id, conversation_id)

    return list(reversed(events))


async def _post_and_reason(bot_id, conversation_id: str, event: dict):
    """Posts an event and runs the reasoning."""
    return await rasa_api.add_event(bot_id, conversation_id, event)


# noinspection PyPep8Naming
@app.post("/twilio/voice/{bot_id}")
async def twilio_voice(bot_id: str,
                       request: Request,
                       CallSid: str = Form(),
                       CallStatus: Optional[str] = Form(""),
                       From: Optional[str] = Form(""),
                       To: Optional[str] = Form(""),
                       SpeechResult: Optional[str] = Form(""),
                       Digits: Optional[str] = Form("")):
    conversation_id = CallSid
    t0 = time()
    log.info(f"Got new Twilio call for conversation {conversation_id}")
    log.info(request.url)

    # First, we need to determine if it's a new call
    tracker, initial_events = await rasa_api.get_tracker_and_events(bot_id, conversation_id)

    twilio_config = tracker["slots"].get("twilio")
    voice = twilio_config["voice"] or "Polly.Joanna-Neural"
    lang = twilio_config["lang"] or "en-US"

    # If we have only 1 event (which should be action_listen), it's a new call
    new_call = False
    if len(initial_events) == 1:
        # So, we post the new call event
        initial_events = await _post_and_reason(bot_id, conversation_id, {"event_type": "new_conversation"})
        new_call = True

    response_elements = []

    if CallStatus == "completed":
        await _post_and_reason(bot_id, conversation_id, {"event_type": "call ended"})
    else:
        if new_call:
            events = await _post_and_reason(bot_id, conversation_id, {
                "event_type": "new call",
                "event_data": {
                    "phone_number": From,
                    "called_phone_number": To
                }
            })
        else:
            _input = SpeechResult or Digits

            if _input:
                log.info(f"The user input was: '{_input}'")
                events = await _post_and_reason(bot_id, conversation_id, {
                    "event_type": "user_said",
                    "event_data": {
                        "value": _input
                    }
                })
            else:
                log.info("The user was silent.")
                events = await _post_and_reason(bot_id, conversation_id, {
                    "event_type": "user_silent",
                    "event_data": {}
                })

        new_events = events[len(initial_events):]

        for event in new_events:
            if event["event_type"] == "bot_said":
                text = event["event_data"]["value"]

                response_elements.append(f"""
                    <Say voice="{voice}" language="{lang}">
                                {text}
                            </Say>
                """)
            elif event["event_type"] == "action" and event["event_data"]["action"] == "hang_up":
                response_elements.append(f"""<Hangup/>""")

        # Last, we need to add the gather element
        defaults = {
            "input": "speech dtmf",
            "enhanced": "true",
            "hints": "",
            "timeout": 5,
            "speechTimeout": "auto",
            "speechModel": "phone_call",
            "language": lang
        }

        response_elements.append(f"""
            <Gather input="{defaults['input']}" 
                            enhanced="{defaults['enhanced']}" 
                            speechTimeout="{defaults['speechTimeout']}" 
                            speechModel="{defaults['speechModel']}"
                            timeout="{defaults['timeout']}"
                            hints="{defaults['hints']}"
                            actionOnEmptyResult="true" 
                            action="/twilio/voice/{bot_id}#e=dublin"
                            language="{defaults['language']}"/>
        """)

    response_elements_text = "\n".join(response_elements)
    data = f"""<Response>
        {response_elements_text}
        </Response>
        """

    log_data = data.replace("\n", " ")
    log.info(f"Twilio response in {time() - t0} ms:\n{log_data}")

    # If it's a new call, we also kick off the recording
    if new_call and twilio_config.get("call_recording"):
        loop = asyncio.get_running_loop()
        loop.create_task(start_call_recording(bot_id, conversation_id))

    return Response(content=data, media_type="application/xml")


async def start_call_recording(bot_id: str, conversation_id: str):
    loop = asyncio.get_running_loop()
    recording_id = await loop.run_in_executor(None, start_recording, conversation_id)

    await _post_and_reason(bot_id, conversation_id, {
        "event_type": "call recording started",
        "event_data": {
            "call_id": conversation_id,
            "call_recording_id": recording_id
        }
    })
    log.info(f"Successfully saved call recording in context {recording_id}.")


@app.get("/webchat/{bot_id}")
async def webchat(bot_id: str, auth_key: str = Depends(get_auth_key)):
    log.info(f"[{auth_key}] Got request for webchat for bot {bot_id}")

    data = f"""
<!DOCTYPE html>
<html>

<head>
  <meta charset="UTF-8" />
  <meta name="viewport" content="width=device-width, initial-scale=1.0">
  <title>WebChat</title>

  <style>
    body {{
      color: #ffffff !important;
      padding: 0;
      margin: 0;
    }}

    .content {{
      padding-top: 200px;
      max-width: 60%;
      padding-right: 150px;
      margin: auto;
      color: black;
    }}
  </style>

</head>

<body>
  <div class="cover"></div>
  <div class="content">
    <div id="webchat"
         data-widget="web-chat"
         data-prop-api-url="/"
         data-prop-api-key-id="webchat"
         data-prop-api-key-secret="{WEBCHAT_KEY}"
         data-prop-assistant-id="{bot_id}"
         data-prop-assistant-name="WebChat"
         data-prop-welcome-message=""

         data-prop-assistant-avatar=""
         data-prop-assistant-avatar-margin-top=""
         data-prop-logo-size=""
         data-prop-color=""

         data-prop-auto-toggle="true"
         data-prop-auto-toggle-delay="500"
         data-prop-message-rate-delay="1000"
         data-prop-width="400"
         data-prop-position="inline"
         data-prop-embedded="false"

         data-prop-live-log-mode="test"
         data-prop-welcome-message="Hey there!"
         data-prop-persist-session="false"
         data-prop-fullscreen="false"/>
  </div>
  <div></div>
  <script type="application/javascript">
    let elem = document.getElementById('webchat');
    elem.setAttribute("data-prop-api-url", window.location.origin);
  </script>
  <script src="https://static.colang.ai/webchat.js"></script>
</body>
</html>
"""

    return Response(content=data, media_type="text/html")



import random
import re

from .default_functions import default_functions
from rasa_sdk import Tracker


class AttributeDict(dict):
    """A simple helper class to allow for dot notation when accessing dict members."""
    def __getattr__(self, attr):
        val = self.get(attr, None)
        if isinstance(val, dict):
            return AttributeDict(val)
        elif isinstance(val, list) and len(val) > 0 and isinstance(val[0], dict):
            return [AttributeDict(x) for x in val]
        else:
            return val

    def __setattr__(self, attr, value):
        self[attr] = value


def eval_expr(tracker: Tracker, expr: str) -> any:
    """Evaluates the given expression in the context of the tracker.

    Args:
        tracker: The tracker to evaluate the expression in.
        expr: The expression to evaluate.
    """
    if not expr:
        return None

    # Hacked support for "no" instead of "not", "is" and "is not"
    expr = re.sub(r"(^| )no ", " not ", expr)
    expr = re.sub(r" is ", " == ", expr)
    expr = re.sub(r" is not ", " != ", expr)

    # First, we extract all the variable names and replace $ with "var_.
    var_names = re.findall(r'(?<!")\$([a-zA-Z_][a-zA-Z0-9_]*)', expr)
    updated_expr = re.sub(r'(?<!")\$([a-zA-Z_][a-zA-Z0-9_]*)', r"var_\1", expr)
    expr_locals = {}

    for var_name in var_names:
        # if we've already computed the value, we skip
        if f"var_{var_name}" in expr_locals:
            continue

        val = tracker.slots.get(var_name, None)
        if isinstance(val, dict):
            val = AttributeDict(val)

        expr_locals[f"var_{var_name}"] = val

    # Next, we just evaluate the expression using python.
    try:
        return eval(
            updated_expr, {"random": random, **expr_locals, **default_functions}
        )
    except Exception as ex:
        raise Exception(f"Error evaluating '{expr}' ('{updated_expr}'): {str(ex)}")



import logging
import os
import time

from twilio.rest import Client

log = logging.getLogger(__name__)


def start_recording(conversation_id: str):
    """Starts the recording for the specified conversation.

    We start with a small delay to allow for the resource to be properly created.
    """
    time.sleep(0.25)

    twilio_account_sid = os.environ["TWILIO_ACCOUNT_SID"]
    twilio_auth_token = os.environ["TWILIO_AUTH_TOKEN"]
    twilio_client = Client(twilio_account_sid, twilio_auth_token)

    recording = twilio_client.calls(conversation_id).recordings.create()

    log.info(f"Call recording for {conversation_id} started.")
    return recording.sid



from rasa_sdk.executor import ActionExecutor

from .core import InitializeConversation, Eval, Set, Push, Pop, CreateEvent
from .helpers import action, ActionContext
from .patch_action_executor import run


def init_corasa_actions():
    # Monkey matching of the ActionExecutor class.
    ActionExecutor.run = run




"""Wrapper for the RASA API."""
import json
import logging
import os
import uuid
from typing import Optional

import aiohttp

from .utils import rasa_events_to_co_events

log = logging.getLogger(__name__)

# The URLs on which the RASA servers are listening.
RASA_SERVERS = os.environ.get("RASA_SERVERS", "default=127.0.0.1:5005")
BOT_API_URL = dict([(item.split("=")[0], item.split("=")[1])
                    for item in RASA_SERVERS.split(",")])
BOT_API_URL["default"] = "127.0.0.1:5005"

print(f"RASA Servers: {json.dumps(BOT_API_URL)}")


def _get_bot_api_url(bot_id: str):
    api_url = BOT_API_URL.get(bot_id, BOT_API_URL["default"])
    if not api_url.startswith("http"):
        api_url = f"http://{api_url}"

    return api_url


async def rasa_api_get(bot_id: str, endpoint: str):
    """Performs a GET call to the RASA HTTP server."""
    api_url = _get_bot_api_url(bot_id)

    async with aiohttp.ClientSession() as session:
        async with session.get(f'{api_url}{endpoint}') as response:
            return await response.json()


async def rasa_api_post(bot_id: str, endpoint: str, data: Optional[dict] = None):
    """Performs a POST call to the RASA HTTP server."""
    api_url = _get_bot_api_url(bot_id)

    async with aiohttp.ClientSession() as session:
        url = f'{api_url}{endpoint}'
        async with session.post(url, json=data) as response:
            log.debug(f"RASA call to {url} returned status: {response.status}")

            return await response.json()


def _reinforces(intent_a: str, intent_b: str) -> bool:
    return intent_a in intent_b or intent_b in intent_a


def _update_intent_confidence(result):
    """Computes the right intent confidence based on the data coming from the NLU server.

    The data has the format:
    {
        "text": "",
        "intent": {
           "id": 517191430183595825,
           "name": "express_greeting_hello",
           "confidence": 0.7905938625335693
          },
          "entities": [],
          "intent_ranking": [
           {
            "id": 517191430183595825,
            "name": "express_greeting_hello",
            "confidence": 0.7905938625335693
           },
           ...
           ]
        }
    }
    """
    i = 1
    intent = result["intent_ranking"][0]["name"]

    while i < len(result["intent_ranking"]):
        additional_intent = result["intent_ranking"][i]

        if intent in additional_intent["name"] or additional_intent["name"] in intent:
            print(f"Updated confidence for `{intent}` "
                  f"using {additional_intent['name']}: "
                  f"+{additional_intent['confidence']}")

            result["intent"]["confidence"] += additional_intent["confidence"]

        i += 1


async def _compute_custom_features(bot_id, conversation_id):
    """Computes the custom features at a specific point in the conversation.

    We use the data stored in the `nlu_custom_features` slot and the last `utter_*`.
    """
    tracker = await rasa_api_get(bot_id, f"/conversations/{conversation_id}/tracker")

    last_utter = None
    i = len(tracker["events"])
    while i > 0:
        i -= 1
        event = tracker["events"][i]
        if event["event"] == "action" and event["name"].startswith("utter_"):
            last_utter = event["name"]
            break

    if last_utter is None:
        return None

    nlu_custom_features = tracker["slots"].get("nlu_custom_features", {})

    # We try to identify if there is a custom feature associated with the last utter
    for item in nlu_custom_features.values():
        if last_utter in item["after"]:
            return item["code"]

    return None


async def add_event(bot_id: str, conversation_id: str, event: dict):
    """Adds a new event to the RASA conversation.

    For events different than user_said, we add a synthetic user intent 'event'.
    """
    if "event_data" not in event:
        event = {
            **event,
            "event_data": {}
        }
    if event["event_type"] == "user_said":
        text = event["event_data"]["value"]
        parse_text = text

        # We compute the custom features before parsing
        custom_feature = await _compute_custom_features(bot_id, conversation_id)
        if custom_feature:
            log.info(f"Found custom feature {custom_feature} for {text}")
            parse_text += " " + custom_feature

        # First, we use the NLU endpoint to parse the data
        result = await rasa_api_post(bot_id, f"/model/parse", data={
            "text": parse_text,
        })

        _update_intent_confidence(result)

        # If the confidence is below 0.5, we transform it into "unknown_intent"
        if result["intent"]["confidence"] < 0.5:
            result["intent"]["name"] = "unknown_intent"
            result["intent"]["confidence"] = 1 - result["intent"]["confidence"]
            result["entities"] = []

        # We also add the hardcoded "entities" entity with all the values from
        # the latest intent.
        entities_dict = {}
        for entity in result["entities"]:
            entities_dict[entity["entity"]] = entity["value"]

        result["entities"].append({
            "start": 0,
            "end": 0,
            "entity": "entities",
            "value": entities_dict
        })

        log.info(f"NLU result: {json.dumps(result)}.")

        await rasa_api_post(bot_id, f"/conversations/{conversation_id}/tracker/events", data={
            "event": "user",
            "text": text,
            "parse_data": {
                "intent": result["intent"],
                "entities": result["entities"],
                "text": text
            }
        })
    else:
        # Create the synthetic intent.
        text = json.dumps(event)

        # For events, we only set the `event` entity.
        entities = [{
            "start": 0,
            "end": 0,
            "entity": "event_data",
            "value": event["event_data"]
        }]

        await rasa_api_post(bot_id, f"/conversations/{conversation_id}/tracker/events", data={
            "event": "user",
            "text": text,
            "parse_data": {
                "intent": {
                    "name": f"event_{event['event_type']}",
                    "confidence": 1.0
                },
                "entities": entities,
                "text": text
            }
        })

    events = await run_reasoning(bot_id, conversation_id)

    return events


async def create_conversation(bot_id: str, conversation_data: dict):
    """Creates a new conversation using the RASA API.

    We do this by posting the "new_conversation" event.
    """
    conversation_id = conversation_data.get("conversation_id")

    if not conversation_id:
        conversation_id = str(uuid.uuid4())

    log.info(f"Creating new conversation {conversation_id}")
    events = await add_event(bot_id, conversation_id, {
        "event_type": "new_conversation",
        "event_data": conversation_data
    })

    return conversation_id, events


async def list_events(bot_id: str, conversation_id: str):
    """Returns the list of events for the given conversation."""
    tracker = await rasa_api_get(bot_id, f"/conversations/{conversation_id}/tracker")

    return rasa_events_to_co_events(tracker["events"], tracker)


async def get_tracker_and_events(bot_id: str, conversation_id: str):
    """Returns the full tracker for the given conversation."""
    tracker = await rasa_api_get(bot_id, f"/conversations/{conversation_id}/tracker")
    events = rasa_events_to_co_events(tracker["events"], tracker)

    return tracker, events


async def run_reasoning(bot_id: str, conversation_id: str):
    """Predicts and executes the next action until action_listen."""
    log.info(f"Running reasoning for conversation {conversation_id}")
    while True:
        res = await rasa_api_post(bot_id, f"/conversations/{conversation_id}/predict")

        next_action = res["scores"][0]["action"]
        log.info(f"Next action is {next_action}")

        # At action_listen, we check if we need to live log
        if next_action == "action_listen":
            # TODO: optimize this?
            tracker = await rasa_api_get(bot_id, f"/conversations/{conversation_id}/tracker")
            if tracker["slots"].get("live_log") or False:
                await rasa_api_post(bot_id, f"/conversations/{conversation_id}/execute", data={
                    "name": "live_log",
                    "policy": res["policy"],
                    "confidence": 1.0
                })
                await rasa_api_post(bot_id, f"/conversations/{conversation_id}/tracker/events", data={
                    "event": "undo"
                })

        await rasa_api_post(bot_id, f"/conversations/{conversation_id}/execute", data={
            "name": next_action,
            "policy": res["policy"],
            "confidence": res["confidence"]
        })

        if next_action == "action_listen":
            break

    _, events = await get_tracker_and_events(bot_id, conversation_id)
    return events



import logging
import re
from typing import List, Dict, Text, Optional

from rasa.core.nlg import interpolator
from rasa.core.policies.rule_policy import RULES, RulePolicy
from rasa.nlu.components import Component
from rasa.shared.core.generator import TrackerWithCachedStates
from .datetime_util import nlg_formatter

logger = logging.getLogger(__name__)


class MonkeyPatching(Component):
    """Patches a few things to improve the RASA behavior."""
    pass


def interpolate_text(response: Text, values: Dict[Text, Text]) -> Text:
    """Interpolate values into responses with placeholders.

    Transform response tags from "{tag_name}" to "{0[tag_name]}" as described here:
    https://stackoverflow.com/questions/7934620/python-dots-in-the-name-of-variable-in-a-format-string#comment9695339_7934969
    Block characters, making sure not to allow:
    (a) newline in slot name
    (b) { or } in slot name

    Args:
        response: The piece of text that should be interpolated.
        values: A dictionary of keys and the values that those
            keys should be replaced with.

    Returns:
        The piece of text with any replacements made.
    """
    try:
        # print(f"Interpolating {response}.")

        # First, we compute the filters for every value that needs to be interpolated
        matches = re.findall(r"{([^}]+)}\|(\w+)", response)
        filters = {}
        for key, filter_name in matches:
            filters[key] = filter_name

        # Next, we get rid of the filters
        response = re.sub(r"}\|\w+", "}", response)

        # Next, we apply the filters to all keys present in the response
        matches = re.findall(r"{([^}]+)}", response)
        for key in matches:
            key_filters = [filters[key]] if key in filters else []
            obj = values
            while "." in key:
                _start, key = key.split(".", 1)
                obj = values.get(_start, {})

            obj[key] = nlg_formatter(obj.get(key, ""), key_filters)

            # print(f"Adjusted '{key}' to '{obj[key]}'")

        text = re.sub(r"{([^\n{}]+?)}", r"{0[\1]}", response)
        text = text.format(values)
        if "0[" in text:
            # regex replaced tag but format did not replace
            # likely cause would be that tag name was enclosed
            # in double curly and format func simply escaped it.
            # we don't want to return {0[SLOTNAME]} thus
            # restoring original value with { being escaped.
            return response.format({})

        return text
    except KeyError as e:
        logger.exception(
            f"Failed to replace placeholders in response '{response}'. "
            f"Tried to replace '{e.args[0]}' but could not find "
            f"a value for it. There is no slot with this "
            f"name nor did you pass the value explicitly "
            f"when calling the response. Return response "
            f"without filling the response. "
        )
        return response


interpolator.interpolate_text = interpolate_text


def _check_prediction(
        self,
        tracker: TrackerWithCachedStates,
        predicted_action_name: Optional[Text],
        gold_action_name: Text,
        prediction_source: Optional[Text],
) -> List[Text]:
    if not predicted_action_name or predicted_action_name == gold_action_name:
        return []

    # Also, disable checking when the predicted action name is "action_listen".
    # This is because of conflicts caused by implicitly adding "listen" at the end of
    # the stories.
    if gold_action_name == "action_listen":
        return []

    if self._should_delete(prediction_source, tracker, predicted_action_name):
        self.lookup[RULES].pop(prediction_source)
        return []

    tracker_type = "rule" if tracker.is_rule_tracker else "story"
    contradicting_rules = {
        rule_name
        for rule_name, action_name in self._rules_sources[prediction_source]
        if action_name != gold_action_name
    }

    error_message = (
        f"- the prediction of the action '{gold_action_name}' in {tracker_type} "
        f"'{tracker.sender_id}' "
        f"is contradicting with rule(s) '{', '.join(contradicting_rules)}'"
    )
    # outputting predicted action 'action_default_fallback' is confusing
    if predicted_action_name != self._fallback_action_name:
        error_message += f" which predicted action '{predicted_action_name}'"

    return [error_message + "."]


RulePolicy._check_prediction = _check_prediction


print("CoRASA monkey patching applied.")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['corasa', 'corasa.actions', 'corasa.api', 'corasa.components']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp==3.8.1',
 'fastapi>=0.79.0,<0.80.0',
 'pydantic==1.9.1',
 'python-dateutil==2.8.1',
 'python-multipart==0.0.5',
 'requests>=2.24.0,<3.0.0',
 'starlette==0.19.1',
 'twilio==6.50.1',
 'uvicorn==0.18.2']

setup_kwargs = {
    'name': 'corasa',
    'version': '0.1.17',
    'description': 'Components for running a RASA project exported from Colang.',
    'long_description': None,
    'author': 'Razvan Dinu',
    'author_email': 'drazvan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from datetime import datetime
from typing import Optional, List


DEFAULT_CHANNEL_EVENTS = [
    "new_conversation",
    "conversation_initialized",
    "new call",
    "call ended"
    "user_online",
]

DEFAULT_CHANNEL_ACTIONS = [
    "hang_up",
    "dial"
]


def rasa_event_to_co_event(e: dict, channel_events: Optional[List[str]] = None):
    """Transforms an event in RASA format to the standard co format."""
    # First, we deal with the id and timestamp.
    dt = datetime.fromtimestamp(e["timestamp"])

    event = {
        "event_id": str(e["timestamp"]).replace(".", ""),
        "event_type": "",
        "event_data": {},
        "created_at": dt.isoformat()
    }

    if e["event"] == "user":
        # We need to look at parse data to determine if it's a synthetic event or not
        if e["parse_data"]["intent"]["name"].startswith("event_"):
            event_name = e["parse_data"]["intent"]["name"][6:]

            # We only transform the event if it's a channel one
            if event_name in DEFAULT_CHANNEL_EVENTS or event_name in (channel_events or []):
                event["event_type"] = event_name
                for entity in e["parse_data"]["entities"]:
                    event["event_data"].update({
                        entity["entity"]: entity["value"]
                    })
        else:
            event["event_type"] = "user_said"
            event["event_data"] = {
                "value": e["text"]
            }
    elif e["event"] == "bot":
        event["event_type"] = "bot_said"
        event["event_data"] = {
            "value": e["text"]
        }
    elif e["event"] == "action" and e["name"] == "action_listen":
        event["event_type"] = "action"
        event["event_data"] = {
            "action": "listen",
            "params": {}
        }
    elif e["event"] == "action" and e["name"] in DEFAULT_CHANNEL_ACTIONS:
        event["event_type"] = "action"

        # TODO: extract params if present
        event["event_data"] = {
            "action": e["name"],
            "params": {}
        }

    return event if event["event_type"] else None


def rasa_events_to_co_events(events: list, tracker: dict):
    """Transforms a full history of RASA events into co events.

    Events that don't have an equivalent are skipped.
    """
    channel_events = tracker["slots"].get("channel_events", [])
    co_events = []
    for event in events:
        co_event = rasa_event_to_co_event(event, channel_events)
        if co_event:
            co_events.append(co_event)

    return co_events
